import json

import pytest
import pyvfg
from pydantic import BaseModel, ValidationError
from pyvfg import VFGPydanticType


# Create a model class that uses this annotation as a field
class VFGModel(BaseModel):
    vfg: VFGPydanticType


@pytest.fixture
def sprinkler_vfg() -> str:
    with open("./test_data/small/sprinkler_factor_graph_vfg.json", "r") as f:
        return f.read()


def test_model_passthrough(sprinkler_vfg):
    vfg = pyvfg.vfg_from_json(sprinkler_vfg)
    model = VFGModel(vfg=vfg)


def test_model_from_json_str(sprinkler_vfg):
    model = VFGModel(vfg=sprinkler_vfg)


def test_model_from_json_object(sprinkler_vfg):
    model = VFGModel(vfg=json.loads(sprinkler_vfg))


def test_model_from_str_invalid():
    with pytest.raises(ValidationError):
        model = VFGModel(vfg="test")


def test_model_from_unsupported_python_type_invalid():
    with pytest.raises(ValidationError):
        model = VFGModel(vfg="test")


def test_model_dump(sprinkler_vfg):
    model = VFGModel(vfg=sprinkler_vfg)
    print(model.model_dump())


def test_model_sdump_json(sprinkler_vfg):
    model = VFGModel(vfg=sprinkler_vfg)
    print(model.model_dump_json())


def test_model_json_schema(sprinkler_vfg):
    model = VFGModel(vfg=sprinkler_vfg)
    print(json.dumps(model.model_json_schema()))



# -*- coding: utf-8 -*-

import numpy as np
import pytest
import pyvfg


@pytest.fixture
def sprinkler_vfg() -> str:
    with open("./test_data/small/sprinkler_factor_graph_vfg.json", "r") as f:
        return f.read()


def test_vfg_from_json(sprinkler_vfg):
    graph = pyvfg.vfg_from_json(sprinkler_vfg)

    # check graph
    assert graph is not None, "Graph should be created from JSON"
    assert len(graph.variables) == 4, "Sprinkler graph should have 4 variables"
    assert graph.version == "0.3.0", "Graph version should be the latest version"


def test_vfg_to_json(sprinkler_vfg):
    import json

    tmp_graph = pyvfg.vfg_from_json(sprinkler_vfg)
    graph_json2 = pyvfg.vfg_to_json(tmp_graph)

    graph = json.loads(sprinkler_vfg)
    graph2 = json.loads(graph_json2)

    assert len(graph["variables"]) == len(
        graph2["variables"]
    ), "Setting the graph lets us get the same variable count"
    assert len(graph["factors"]) == len(
        graph2["factors"]
    ), "Setting the graph gets us the same factor count"


def test_set_graph(sprinkler_vfg):
    graph = pyvfg.vfg_from_json(sprinkler_vfg)
    pyvfg.set_graph(graph)
    graph2 = pyvfg.get_graph()

    assert graph == graph2, "Setting the graph lets us get the same graph"


def test_probability_distribution():
    assert (
        pyvfg.ProbabilityDistribution.Categorical
        == pyvfg.ProbabilityDistribution.Categorical
    )
    assert (
        pyvfg.ProbabilityDistribution.CategoricalConditional
        == pyvfg.ProbabilityDistribution.CategoricalConditional
    )

    assert (
        pyvfg.ProbabilityDistribution.Categorical
        != pyvfg.ProbabilityDistribution.CategoricalConditional
    )

    assert int(pyvfg.ProbabilityDistribution.Categorical) == 0
    assert int(pyvfg.ProbabilityDistribution.CategoricalConditional) == 1


def test_role():
    assert pyvfg.FactorRole.Transition == pyvfg.FactorRole.Transition
    assert pyvfg.FactorRole.Preference == pyvfg.FactorRole.Preference
    assert pyvfg.FactorRole.Likelihood == pyvfg.FactorRole.Likelihood
    assert pyvfg.FactorRole.InitialStatePrior == pyvfg.FactorRole.InitialStatePrior

    assert pyvfg.FactorRole.Transition != pyvfg.FactorRole.Preference
    assert pyvfg.FactorRole.Transition != pyvfg.FactorRole.Likelihood
    assert pyvfg.FactorRole.Preference != pyvfg.FactorRole.Likelihood
    assert pyvfg.FactorRole.Likelihood != pyvfg.FactorRole.InitialStatePrior

    assert int(pyvfg.FactorRole.Transition) == 1
    assert int(pyvfg.FactorRole.Preference) == 2
    assert int(pyvfg.FactorRole.Likelihood) == 3
    assert int(pyvfg.FactorRole.InitialStatePrior) == 4


def test_create_vfg():
    graph = pyvfg.VFG(
        factors=[
            pyvfg.Factor(
                variables=["rain"],
                distribution=pyvfg.ProbabilityDistribution.Categorical,
            )
        ],
        variables={"rain": pyvfg.DiscreteVariableNamedElements(["on", "off"], None)}
    )
    assert graph != pyvfg.VFG.default()

def test_nice_names():
    assert "%s" % (type(pyvfg.FactorRole.Transition)) == "<class 'pyvfg.FactorRole'>"
    assert "%s" % (type(pyvfg.VFG.default())) == "<class 'pyvfg.VFG'>"
    assert "%s" % (type(pyvfg.Factor.default())) == "<class 'pyvfg.Factor'>"
    assert "%s" % (type(pyvfg.DiscreteVariableNamedElements([]))) == "<class 'pyvfg.DiscreteVariableNamedElements'>"
    assert "%s" % (type(pyvfg.DiscreteVariableAnonymousElements(0))) == "<class 'pyvfg.DiscreteVariableAnonymousElements'>"
    assert "%s" % (type(pyvfg.ProbabilityDistribution.Categorical)) == "<class 'pyvfg.ProbabilityDistribution'>"


def test_factor(sprinkler_vfg):
    graph = pyvfg.vfg_from_json(sprinkler_vfg)
    factors = graph.factors

    assert isinstance(factors, list)

    for factor in factors:
        assert isinstance(factor.variables, list)
        assert isinstance(factor.distribution, pyvfg.ProbabilityDistribution)
        assert isinstance(factor.values, np.ndarray)
        assert factor.role == pyvfg.FactorRole.NoRole


def test_vfg(sprinkler_vfg):
    graph = pyvfg.vfg_from_json(sprinkler_vfg)

    assert isinstance(graph.version, str)
    assert isinstance(graph.factors, list)
    assert isinstance(graph.variables, dict)

def test_validate(sprinkler_vfg):
    graph = pyvfg.vfg_from_json(sprinkler_vfg)
    pyvfg.validate_graph(graph)

def test_validate_failure():
    with pytest.raises(pyvfg.JsonSerializationError):
        pyvfg.vfg_from_json("not_json")
    with pytest.raises(pyvfg.JsonSerializationError):
        pyvfg.vfg_from_json("{}")

def test_variable_export(sprinkler_vfg):
    """
    Tests that our Variable type is properly exported to python
    :param sprinkler_vfg: from fixture
    :return: None
    """
    graph = pyvfg.vfg_from_json(sprinkler_vfg)
    rain_var = graph.variables["rain"]
    match type(rain_var):
        case pyvfg.DiscreteVariableNamedElements:
            assert isinstance(rain_var.elements, list)
            assert isinstance(rain_var.role, pyvfg.VariableRole)
        case _:
            assert False, "Should be a discrete variable with named elements"


def test_types_export(sprinkler_vfg):
    """
    Tests that the types exactly align with what we expect.
    :param sprinkler_vfg: from fixture
    :return: None
    """
    vfg = pyvfg.vfg_from_json(sprinkler_vfg)
    assert type(vfg) == pyvfg.VFG, "root type should be pyvfg.VFG"
    assert type(vfg.factors[0]) == pyvfg.Factor, "Factor type should be pyvfg.Factor"
    assert type(vfg.variables["cloudy"]) == pyvfg.DiscreteVariableNamedElements, "Variable type should be pyvfg.DiscreteVariableNamedElements"

def test_version_export() -> None:
    """
    Tests that __version__ is exported
    :return: None
    """
    import re

    assert hasattr(pyvfg, "__version__")
    assert isinstance(pyvfg.__version__, str)
    re.match("\\d+\\.\\d+\\.\\d+", pyvfg.__version__)



# -*- coding: utf-8 -*-
import typing

# this is an interface file for python tooling. values are declared here and defined in `src/python/mod.rs`.

__version__ = ...

class VFGException(Exception):
    ...

@typing.final
class FileManipulationError(VFGException):
    ...

@typing.final
class RkyvDeserializationError(VFGException):
    ...

@typing.final
class JsonSerializationError(VFGException):
    ...

@typing.final
class ValidationError(VFGException):
    ...

@typing.final
class InvalidGraphVersionError(VFGException):
    ...

@typing.final
class ProbabilityDistribution:
    """
    Probability distribution.
    """

    Categorical: int = 0
    CategoricalConditional: int = 1

    def __richcmp__(self, other: ProbabilityDistribution, op) -> bool: ...
    def __int__(self) -> int: ...

@typing.final
class VariableRole:
    """
    VariableRole can be one of three values: NoRole, ControlState, or Latent.
    """
    NoRole: int = 0
    ControlState: int = 1
    Latent: int = 2

    def __richcmp__(self, other: VariableRole, op) -> bool: ...
    def __int__(self) -> int: ...

@typing.final
class DiscreteVariableNamedElements:
    """
    DiscreteVariableNamedElements represents a discrete variable with named elements.
    """

    elements: list[str]
    role: VariableRole

    def __init__(self, elements: list[str], role: typing.Optional[VariableRole] = None) -> None: ...

@typing.final
class DiscreteVariableAnonymousElements:
    """
    DiscreteVariableAnonymousElements represents a discrete variable with anonymous elements.
    """

    cardinality: int
    role: VariableRole

    def __init__(self, cardinality: int, role: typing.Optional[VariableRole] = None) -> None: ...

@typing.final
class FactorRole:
    """
    FactorRole is optional can can be one of 3 values: "transition", "preference" or "likelihood".
    There is no default value, only if specified on the factor will it exist
    None is used for the default value in the event that it exists and the numeric value doesn't match the enum
    """
    NoRole: int = 0
    Transition: int = 1
    Preference: int = 2
    Likelihood: int = 3
    InitialStatePrior: int = 4

    def __richcmp__(self, other: FactorRole, op) -> bool: ...
    def __int__(self) -> int: ...

@typing.final
class Factor:
    """
    A Factor represents a single factor extraction of the factor graph.
    It contains variables, values, and a type.
    """

    variables: list[str]
    distribution: ProbabilityDistribution
    """ values is a numpy array """
    values: list[typing.Any]
    role: FactorRole

    @staticmethod
    def default() -> Factor: ...

    def __init__(self, variables: list[str], distribution: typing.Union[ProbabilityDistribution, int], role: typing.Optional[FactorRole] = None) -> None: ...

@typing.final
class VFG:
    """
    Represents the entire VFG.
    A VFG consists of a list of factors, and a map of variables to their metadata.
    For discrete variables, the only metadata is a list of their possible values.

    :param version:
    :param factors:
    :param variables:
    """

    version: str
    factors: list[Factor]
    variables: dict[str, typing.Union[DiscreteVariableNamedElements, DiscreteVariableAnonymousElements]]

    @staticmethod
    def default() -> VFG: ...

    def __init__(self, factors: list[Factor], variables: dict[str, typing.Union[DiscreteVariableNamedElements, DiscreteVariableAnonymousElements]]) -> None: ...

def get_graph() -> VFG:
    """
    Get the global VFG.
    """
    ...

def set_graph(new_graph: VFG) -> None:
    """
    Set the global VFG.

    :param new_graph: The graph to set
    """
    ...

def validate_graph(graph: VFG) -> None:
    """
    Validates VFG.

    :param graph: The graph to validate
    """
    ...

def get_subgraph_from(variable_name: list[str]) -> VFG:
    """
    Retrieve subgraph.

    :param variable_name:
    """
    ...

def vfg_to_json(vfg: VFG) -> str:
    """
    Serialize a VFG to a JSON string.

    :param vfg:
    """
    ...

def vfg_from_json(json: str) -> VFG:
    """
    Deserialize a JSON string to VFG.

    :param json:
    """
    ...



import json
from typing import Annotated, Any

import requests
from pydantic import Field, GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from pyvfg import VFG, vfg_from_json, vfg_to_json

VFG_SCHEMA_REF_URL = "https://verses-json-schema-registry-gitops-usw1.s3.us-west-1.amazonaws.com/vfg/0_3_0/vfg_0_3_0.json"
VFG_EXAMPLE_REF_URL = "https://verses-json-schema-registry-gitops-usw1.s3.us-west-1.amazonaws.com/vfg/examples/small/sprinkler_factor_graph_vfg.json"

VFG_EXAMPLE_JSON = requests.get(VFG_EXAMPLE_REF_URL).json()


class _VFGPydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """
        We return a pydantic_core.CoreSchema that behaves in the following ways:

        * JSON strings will be parsed as `VFG` instances by calling `vfg_from_json()`
        * JSON dictionaries will be parsed as `VFG` instances by calling `vfg_from_json()`
        * `VFG` instances will be parsed as `VFG` instances without any changes
        * Nothing else will pass validation
        * Serialization will always return the string from `vfg_to_json()`
        """

        def validate_from_str(value: str) -> VFG:
            # Pydantic will catch the `ValueError` and raise `ValidationError`
            json.loads(value)
            return vfg_from_json(value)

        def validate_from_dict(value: dict[str, Any]) -> VFG:
            return vfg_from_json(json.dumps(value))

        def serialize(instance: VFG) -> str:
            return json.loads(vfg_to_json(instance))

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        from_dict_schema = core_schema.chain_schema(
            [
                core_schema.dict_schema(),
                core_schema.no_info_plain_validator_function(validate_from_dict),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_dict_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(VFG),
                    from_str_schema,
                    from_dict_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(serialize),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """
        Return an external reference to the VFG schema we currently support.
        Using an external URL allows us to bypass pydantic's internal core_schema checking during
        JSON schema retrieval, which would otherwise throw a `KeyError` if we tried to supply
        the JSON schema as an object directly in this method. This allows the consumer of the
        JSON schema (e.g. FastAPI) to fetch the schema on the client side during API doc
        generation.
        """
        return {
            "$ref": VFG_SCHEMA_REF_URL,
        }


VFGPydanticType = Annotated[
    VFG,
    _VFGPydanticAnnotation,
    Field(
        ...,
        examples=[VFG_EXAMPLE_JSON],
    ),
]
